
package Chat;

import java.io.*;
import java.net.*;
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import sun.misc.*;

/**
 * The interface which defines a console for worker classes.
 * Implementing classes enable the application to run both
 * in text and graphic mode.
 *
 * @see TextConsole
 * @see AWTConsole
 */
interface Console {
  /** Reads a line from the console, blocks until it is available.
   *  @return    the line read.
   */
  String readLine();
  /** Writes a line to the console.
   *  @param     message  the line to be written.
   */
  void writeLine(String message);
  /** Writes an error message to the console.
   *  @param     message  the message to be written.
   */
  void error(String message);
  /** Updates a user list.
   *  @param     list  a collection of <code>String</code> objects.
   */
  void updateUserList(Collection list);
  /** Sets a title of the window.
   *  @param     title  the desired title.
   */
  void setTitle(String title);
  /** Closes the console window.
   *  @param     title  the desired title.
   */
  void close();
}

/**
 * Implementation of text based console. STDIN, STDOUT and STDERR streams
 * are used to interact with the user. 
 *
 * @see AWTConsole
 */
class TextConsole implements Console {
  /** STDIN input stream stored as a <code>BufferedReader</code>. */
  private BufferedReader in;
  /** STDOUT output stream stored as a <code>PrintWriter</code>. */
  private PrintWriter out;
  /** STDERR output stream stored as a <code>PrintWriter</code>. */
  private PrintWriter err;
  /** Initializes member variables with system standard streams.
   */
  TextConsole() {
    in = new BufferedReader(new InputStreamReader(System.in));
    out = new PrintWriter(new OutputStreamWriter(System.out),true);
    err = new PrintWriter(new OutputStreamWriter(System.err),true);
  }
  public String readLine() {
    try {
      return in.readLine();
    }
    catch(IOException e) {
      error("TextConsole failed (" + e.getMessage() + ")");
      return "";
    }
  }
  public void writeLine(String message) {
    out.println(message);
  }
  public void error(String message) {
    err.println(message);
  }
  /** Implementation is empty in this class. */
  public void updateUserList(Collection list) { }
  /** Implementation is empty in this class. */
  public void setTitle(String title) { }
  /** Implementation is empty in this class. */
  public void close() { }
}

/**
 * A class derived from <code>Frame</code> implementing the
 * <code>WindowListener</code> interface by empty methods.
 *
 * @see java.awt.Frame
 * @see java.awt.event.WindowListener
 */
class BasicWindow extends Frame implements WindowListener {
  public void windowClosing(WindowEvent e) { }
  public void windowOpened(WindowEvent e) { }
  public void windowClosed(WindowEvent e) { }
  public void windowIconified(WindowEvent e) { }
  public void windowDeiconified(WindowEvent e) { }
  public void windowActivated(WindowEvent e) { }
  public void windowDeactivated(WindowEvent e) { }
}

/**
 * An abstract class, base for simple command dialogs invoked by
 * <code>AWTConsole</code> window.
 *
 * @see AWTConsole
 */
abstract class StandardDialog extends Dialog implements WindowListener, ActionListener, KeyListener {
  /** The window layout used by the dialog. */
  private GridBagLayout gridbag;
  /** Creates and partly initalizes a dialog.
   *  @param     owner  parent window.
   *  @param     title  window title.
   */
  StandardDialog(Frame owner, String title) {
    super(owner, title, true);
    addWindowListener(this);
    addKeyListener(this);
    gridbag = new GridBagLayout();
    setLayout(gridbag);
    setResizable(false);
  }
  /** Packs the dialog, centers it according to the owner and makes it visible.
   */
  protected void packCenterAndShow() {
    pack();
    Window owner = getOwner();
    setLocation(owner.getX()+(owner.getWidth()-getWidth())/2,owner.getY()+(owner.getHeight()-getHeight())/2);
    setVisible(true);
  }
  /** Adds a file dwith label to a gridbag layout.
   *  @param     label  new Label object.
   *  @param     field  new TextField object.
   *  @return    the passed-in TextField.
   */
  protected TextField addLabeledField(Label label, TextField field) {
    GridBagConstraints c = new GridBagConstraints();
    c.insets = new Insets(5,5,5,5);
    c.anchor = GridBagConstraints.WEST;
    c.gridwidth = 1;
    gridbag.setConstraints(label, c);
    add(label);
    c.gridwidth = GridBagConstraints.REMAINDER;
    gridbag.setConstraints(field, c);
    field.addKeyListener(this);
    add(field);
    return field;
  }
  /** Adds a checkbox to a gridbag layout.
   *  @param     box new CheckBox object.
   *  @return    the passed-in CheckBox.
   */
  protected Checkbox addCheckbox(Checkbox box) {
    GridBagConstraints c = new GridBagConstraints();
    c.insets = new Insets(5,5,5,5);
    c.anchor = GridBagConstraints.WEST;
    c.gridwidth = GridBagConstraints.REMAINDER;
    gridbag.setConstraints(box, c);
    box.addKeyListener(this);
    add(box);
    return box;
  }
  /** Adds default buttons to a gridbag layout, on a separate Panel.
   */
  protected void addOkAndCancel() {
    GridBagConstraints c = new GridBagConstraints();
    c.insets = new Insets(5,5,5,5);
    Panel buttons = new Panel();
    Button okButton = new Button("O.K.");
    okButton.addActionListener(this);
    okButton.addKeyListener(this);
    buttons.add(okButton);
    Button cancelButton = new Button("Cancel");
    cancelButton.addKeyListener(this);
    cancelButton.addActionListener(this);
    buttons.add(cancelButton);
    c.anchor = GridBagConstraints.CENTER;
    c.gridwidth = 2;
    gridbag.setConstraints(buttons, c);
    add(buttons);
  }
  public void windowClosing(WindowEvent e) {
    dispose();
  }
  public void windowOpened(WindowEvent e) { }
  public void windowClosed(WindowEvent e) { }
  public void windowIconified(WindowEvent e) { }
  public void windowDeiconified(WindowEvent e) { }
  public void windowActivated(WindowEvent e) { }
  public void windowDeactivated(WindowEvent e) { }
  /** Checks a key pressed for VK_ENTER and simulates pressing of O.K. button.
   *  @param     e the event.
   */
  public void keyTyped(KeyEvent e) {
    if (e.getKeyChar()==KeyEvent.VK_ENTER)
      actionPerformed(new ActionEvent(this, ActionEvent.ACTION_PERFORMED, "O.K."));
  }
  public void keyPressed(KeyEvent e) { }
  public void keyReleased(KeyEvent e) { }
}

/**
 * Implementation of AWT based console. Console output is represented
 * by <code>TextArea</code> component, user input is taken from
 * <code>TextField</code>. Several buttons are available to invoke
 * console commands. Also features a <code>List</code> of other users.
 *
 * @see TextConsole
 */
class AWTConsole extends BasicWindow implements Console, ActionListener, KeyListener {
  /**
   * A dialog for setting up the local user identity - nickname, host and port.
   */
  class Setup extends StandardDialog {
    /** Stored reference to appropriate textfield. */
    private TextField nickField, hostField, portField;
    /** Holds a returned value after the dialog is dismissed by OK button. */
    String nick, host; int port;
    /** Creates and initalizes the dialog.
     *  @param     owner  parent window.
     */
    Setup(Frame owner) {
      super(owner, "Setup");
      nickField = addLabeledField(new Label("Nick"), new TextField(15));
      hostField = addLabeledField(new Label("Host"), new TextField(Chat.defaultHost));
      portField = addLabeledField(new Label("Port"), new TextField(Integer.toString(Chat.defaultPort)));
      addOkAndCancel();
      packCenterAndShow();
      nickField.requestFocus();
    }
    /** When O.K. button is pressed, stores the field values.
     *  @param     e  the event.
     */
    public void actionPerformed(ActionEvent e) {
      if (e.getActionCommand().equals("Cancel")) {
        dispose();
      }
      else if (e.getActionCommand().equals("O.K.")) {
        try {
          if (nickField.getText().equals("")||hostField.getText().equals(""))
            throw new IllegalArgumentException();
          port = Integer.parseInt(portField.getText());
          nick = nickField.getText();
          host = hostField.getText();
          dispose();
        }
        catch (IllegalArgumentException iae) {
        }
      }
    }
  }
  /**
   * A dialog for connecting to the network.
   */
  class Connect extends StandardDialog {
    /** Stored reference to appropriate textfield. */
    private TextField urlField, roomField, hostField, portField;
    /** A checkbox group representing the two connect methods. */
    private CheckboxGroup methodGroup;
    /** Holds a returned value after the dialog is dismissed by OK button. */
    String url, room, nick, host; int port;
    /** Creates and initalizes the dialog.
     *  @param     owner  parent window.
     */
    Connect(Frame owner) {
      super(owner, "Connect");
      methodGroup = new CheckboxGroup();
      addCheckbox(new Checkbox(" Server", methodGroup, true));
      urlField = addLabeledField(new Label("URL"), new TextField(Chat.defaultURL,35));
      roomField = addLabeledField(new Label("Room"), new TextField(15));
      addCheckbox(new Checkbox(" Direct", methodGroup, false));
      hostField  = addLabeledField(new Label("Host"),  new TextField(15));
      portField  = addLabeledField(new Label("Port"),  new TextField(Integer.toString(Chat.defaultPort)));
      addOkAndCancel();
      packCenterAndShow();
      roomField.requestFocus();
    }
    /** When O.K. button is pressed, stores the field values.
     *  @param     e  the event.
     */
    public void actionPerformed(ActionEvent e) {
      if (e.getActionCommand().equals("Cancel")) {
        dispose();
      }
      else if (e.getActionCommand().equals("O.K.")) {
        try {
          if (methodGroup.getSelectedCheckbox().getLabel().equals(" Server")) {
            if (roomField.getText().equals("")||urlField.getText().equals(""))
              throw new IllegalArgumentException();
            url  = urlField.getText();
            room = roomField.getText();
          }
          else {
            if (hostField.getText().equals(""))
              throw new IllegalArgumentException();
            port = Integer.parseInt(portField.getText());
            host = hostField.getText();
          }
          dispose();
        }
        catch (IllegalArgumentException iae) {
        }
      }
    }
  }
  /** The main text area for received messages. */
  private TextArea textArea;
  /** The input field for outgoing message to be sent. */
  private TextField inputField;
  /** The list of other users on the net. */
  private java.awt.List userList;
  /** A button for appropriate console command. */
  private Button connectButton, disconnectButton, setupButton, clearButton;
  /** Shows whether a message/command is ready to be read by Sender thread . */
  private boolean ready = false;
  /** Stores a line to be read by Sender thread . */
  private String line;
  /** Creates and initalizes the window, places the controls into a GridBag layout.
   */
  AWTConsole() {
    addWindowListener(this);
    GridBagLayout gridbag = new GridBagLayout();
    setLayout(gridbag);

    GridBagConstraints c = new GridBagConstraints();
    c.insets = new Insets(5,5,5,5);

    textArea = new TextArea(10, 40);
    textArea.setEditable(false);
    c.gridx = 0;
    c.gridy = 0;
    c.gridwidth  = 1;
    c.gridheight = 2;
    c.fill = GridBagConstraints.BOTH;
    c.weighty = c.weightx = 1.0;
    gridbag.setConstraints(textArea, c);
    add(textArea);

    userList = new java.awt.List();
    c.gridx = 1;
    c.gridy = 0;
    c.gridwidth = GridBagConstraints.REMAINDER;
    c.gridheight = 1;
    c.fill = GridBagConstraints.BOTH;
    c.weighty = 1.0;
    c.weightx = 0.0;
    gridbag.setConstraints(userList, c);
    add(userList);

    c.gridx = 0;
    c.gridy = 2;
    inputField = new TextField();
    c.gridwidth = c.gridheight = 1;
    c.fill = GridBagConstraints.HORIZONTAL;
    c.weighty = 0.0;
    c.weightx = 1.0;
    gridbag.setConstraints(inputField, c);
    inputField.addKeyListener(this);
    add(inputField);

    c.gridwidth = c.gridheight = 1;
    c.weightx = c.weighty = 0.0;

    c.gridx = 1;  c.gridy = 1;
    clearButton = new Button("Clear");
    clearButton.addActionListener(this);
    gridbag.setConstraints(clearButton, c);
    add(clearButton);

    c.gridx = 2;  c.gridy = 1;
    setupButton = new Button("Setup");
    setupButton.addActionListener(this);
    gridbag.setConstraints(setupButton, c);
    add(setupButton);

    c.gridx = 1;  c.gridy = 2;
    connectButton = new Button("Connect");
    connectButton.addActionListener(this);
    gridbag.setConstraints(connectButton, c);
    add(connectButton);

    c.gridx = 2;  c.gridy = 2;
    disconnectButton = new Button("Disconnect");
    disconnectButton.addActionListener(this);
    gridbag.setConstraints(disconnectButton, c);
    add(disconnectButton);

    pack();
    setIconImage(Toolkit.getDefaultToolkit().getImage("Icon.gif"));
    setIconImage(Toolkit.getDefaultToolkit().getImage(AWTConsole.class.getResource("Icon.gif")));
    setTitle("Chat");
   
    Rectangle bounds = getGraphicsConfiguration().getBounds();
    setLocation(bounds.x+(bounds.width-getWidth())/2,bounds.y+(bounds.height-getHeight())/2);

    setVisible(true);

    inputField.requestFocus();
  }
  /** Tries to passes the string on to a Sender thread, waiting on readLine()
   *  method. If there's no thread waiting, returns with false.
   *  @param     message  a message string.
   *  @return    true when succesful
   *  @see       AWTConsole#readLine()
   */
  private synchronized boolean lineToRead(String message) {
    try {
      if (ready) {
        wait(1000);
        if (ready) return false;
      }
      line = message;
      ready = true;
      notifyAll();
      return true;
    }
    catch (InterruptedException e) {
      return false;
    };
  }
  public void close() {
    System.exit(0);
  }
  /** Reads a line from the console, blocks until it is available.
   *  A call to lineToRead() method causes this method to return.
   *  @return    the line read.
   *  @see       AWTConsole#lineToRead(String message)
   */
  public synchronized String readLine() {
    try {
      while (!ready) wait();
      ready = false;
      notifyAll();
      return line;
    }
    catch (InterruptedException e) {
      return "Interrupted";
    };
  }
  /** Writes an error message to the console - TextArea control.
   *  @param     message  the message to be written.
   */
  public void writeLine(String message) {
    textArea.append(message+"\n");
  }
  /** Writes an error message to the console - TextArea control.
   *  @param     message  the message to be written.
   */
  public void error(String message) {
    writeLine(message);
  }
  public void updateUserList(Collection list) {
    userList.removeAll();
    Iterator i = list.iterator();
    while (i.hasNext()) {
      userList.add((String)i.next());
    }
  }
  public void setTitle(String title) {
    super.setTitle(title);
  }
  /** Handles buttons pressed, invokes dialogs and console commands.
   *  @param     e  the event.
   */
  public void actionPerformed(ActionEvent e) {
    if (e.getActionCommand().equals("Setup")) {
      Setup setup = new Setup(this);
      if (setup.nick!=null) lineToRead("SETUP "+setup.nick+" "+setup.host+" "+setup.port);
    }
    else if (e.getActionCommand().equals("Connect")) {
      Connect connect = new Connect(this);
      if (connect.host!=null) lineToRead("CONNECT "+connect.host+" "+connect.port);
      if (connect.room!=null) lineToRead("ROOM "+connect.url+" "+connect.room);
    }
    else if (e.getActionCommand().equals("Disconnect")) {
      lineToRead("DISCONNECT");
    }
    else if (e.getActionCommand().equals("Clear")) {
      textArea.setText(null);
    }
  }
  public void windowClosing(WindowEvent e) {
    lineToRead("EXIT");
  }
  /** Checks a key pressed for VK_ENTER and sends the entered message out.
   *  @param     e the event.
   */
  public void keyTyped(KeyEvent e) {
    if (e.getKeyChar()==KeyEvent.VK_ENTER) {
      if (lineToRead(inputField.getText())) {
        inputField.setText("");
      }
    }
  }
  public void keyPressed(KeyEvent e) { }
  public void keyReleased(KeyEvent e) { }
}

/**
 * Used by <code>Sender</code> and <code>Receiver</code> to parse both
 * internal and external protocol messages.
 *
 * @see Sender
 * @see Receiver
 */
class MessageParser extends StreamTokenizer {
  /** Stores a name (first string preceding a space) of the message. */
  public final String name;
  /** Initializes the parser with a message string.
   *  @param      string the string to be parsed.
   */
  MessageParser(String string) throws IOException {
    super(new StringReader(string));
    resetSyntax();
    whitespaceChars(' ',' ');
    wordChars('a','z');
    wordChars('A','Z');
    name = (nextToken()==TT_WORD) ? sval : "";
  }
  /** Parses a string token of word characters delimited by space.
   *  @return     the string token.
   */
  String string() throws IOException {
    resetSyntax();
    whitespaceChars(' ',' ');
    wordChars('a','z');
    wordChars('A','Z');
    wordChars('0','9');
    wordChars('.','.');
    if (nextToken()!=TT_WORD) throw new ProtocolException("Wrong message");
    return sval;
  }
  /** Parses an url token delimited by space.
   *  @return     the url.
   */
  String url() throws IOException {
    resetSyntax();
    whitespaceChars(' ',' ');
    wordChars('a','z');
    wordChars('A','Z');
    wordChars('0','9');
    wordChars('.','.');
    wordChars('/','/');
    wordChars(':',':');
    wordChars('?','?');
    wordChars('+','+');
    wordChars('_','_');
    wordChars('-','-');
    wordChars('~','~');
    if (nextToken()!=TT_WORD) throw new ProtocolException("Wrong message");
    return sval;
  }
  /** Parses an integer token delimited by space.
   *  @return     the integer value.
   */
  int integer() throws IOException {
    resetSyntax();
    whitespaceChars(' ',' ');
    parseNumbers();
    if (nextToken()!=TT_NUMBER) throw new ProtocolException("Wrong message");
    return (int)nval;
  }
  /** Parses an remainder of the string.
   *  @return     the end of the string yet unparsed.
   */
  String remainder() throws IOException {
    resetSyntax();
    if (nextToken()!=' ') throw new ProtocolException("Wrong message");
    wordChars(0,255);
    if (nextToken()!=TT_WORD) throw new ProtocolException("Wrong message");
    return sval;
  }
}

/**
 * Maintains a list of other chat users, as well as local user information.
 * Provides a thread-safe simultaneous iterating and updating of the list.
 * <code>Sender</code> and <code>Receiver</code> hold a pointer to the
 * instance of this class.
 *
 * @see Sender
 * @see Receiver
 */
class HostList extends Hashtable {
  /**
   * A data-storage class representing an item in the list of remote users.
   */
  class HostItem {
    /** The nickname of the user. */
    final String nick;
    /** The hostname of the user as a string, either an IP adress or a DNS name. */
    final String host;
    /** The port used by the application. */
    final int port;
    /** Indicates that the object is valid, i.e. enlisted in the <code>HostList</code>. */
    boolean valid;
    /** Constructs an immutable object representing the host list item.
     *  @param     nick user nickname
     *  @param     host user hostname (IP adress)
     *  @param     port user port
     */
    HostItem(String nick, String host, int port) {
      this.nick = nick;
      this.host = host;
      this.port = port;
      valid = false;
    }
    public String toString() {
      return nick+" "+host+" "+port;
    }
  }
  /**
   * A custom thread-safe iterator over the list.
   */
  class HostIterator implements Iterator {
    /** A link to parent <code>HostList</code>. */
    HostList list;
    /** Cached array of list members, used for iterating. */
    Object array[];
    /** An index of the iteration. */
    int i;
    /** Constructs a thread-safe iterator.
     *  @param     list parent <code>HostList</code>
     */
    HostIterator(HostList list) {
      this.list = list;
      array = list.keySet().toArray();
      i = 0;
    }
    public boolean hasNext() {
      return (i<array.length);
    }
    public Object next() {
      return list.get(array[i++]);
    }
    public void remove() {
      throw new UnsupportedOperationException();
    }
  }
  /** A link to the console - used solely for updating the user list. */
  final Console console;
  /** A separately maintained list of nicknames for updating the console user list. */
  final TreeSet nicks;
  /** Information about this (local) user. */
  HostItem me;
  /** Initializes the host list with the link to console.
   *  @param      console link to console.
   */
  HostList(Console console) {
    this.console = console;
    nicks = new TreeSet();
  }
  /** Sets the identity of local user.
   *  @param     nick user nickname
   *  @param     host user hostname (IP adress)
   *  @param     port user port
   */
  void setIdentity(String nick, String host, int port) {
    if (me!=null) nicks.remove(me.nick);
    me = new HostItem(nick, host, port);
    nicks.add(me.nick);
    console.updateUserList(nicks);
  }
  /** Adds a remote user to hostlist.
   *  @param     nick user nickname
   *  @param     host user hostname (IP adress)
   *  @param     port user port
   */
  synchronized void add(String nick, String host, int port) {
    if (!nick.equals(me.nick)) {
      HostItem item = new HostItem(nick, host, port);
      put(nick, item);
      item.valid = true;
      nicks.add(nick);
      console.updateUserList(nicks);
    }
  }
  /** Deletes a remote user from hostlist.
   *  @param     nick user nickname
   */
  synchronized void delete(String nick) {
    HostItem item = (HostItem)remove(nick);
    if (item!=null) item.valid = false;
    nicks.remove(nick);
    console.updateUserList(nicks);
  }
  /** Clears the hostlist.
   */
  public synchronized void clear() {
    super.clear();
    nicks.clear();
    if (me!=null) nicks.add(me.nick);
    console.updateUserList(nicks);
  }
  synchronized Iterator iterator() {
    return new HostIterator(this);
  }
}

/**
 * Encodes/decodes disk files using a BASE64 encoding. This approach to file
 * tranfer is enforced by ascii-based communication protocol of the application. 
 */
class FileTransfer {
  /** A file size limit, in bytes. */
  public static final int MAXSIZE = 1048576;
  /** Reads contents of file specified, encodes it and writes it
   *  out to the <code>out</code> stream.
   *  @param     name  the file name.
   *  @param     out  output characted stream.
   */
  static void encode(String name, PrintWriter out) throws IOException {
    File entry = new File(name);
    int len = (int)entry.length();
    if (len>MAXSIZE) throw new IOException("File too large");
    RandomAccessFile file = new RandomAccessFile(entry,"r");
    byte[] buffer = new byte[len];
    file.readFully(buffer);
    file.close();
    out.println(new BASE64Encoder().encode(buffer));
  }
  /** Decodes the stream <code>in</code> and writes the data to file.
   *  Checks the file size against <code>size</code>.
   *  @param     name  the file name.
   *  @param     in  input characted stream.
   *  @param     size  expected file size.
   */
  static void decode(String name, BufferedReader in, int size) throws IOException {
    CharacterDecoder decoder = new BASE64Decoder();
    OutputStream file = new FileOutputStream(name);
    String line;
    int total = 0;
    while ((line=in.readLine())!=null) {
      byte[] buffer = decoder.decodeBuffer(line);
      file.write(buffer);
      if (buffer.length==0) break;
      total += buffer.length;
    }
    file.close();
    if (total!=size) throw new ProtocolException("File size mismatch");
  }
}

/**
 * A receiver, "server" thread. Listens for protocol messages and processes
 * them sychronously, using <code>Console</code> and <code>HostList</code>
 * objects.
 *
 * @see Sender
 * @see Console
 * @see HostList
 */
class Receiver extends Thread {
  /** The link to common console. */
  private final Console console;
  /** The link to common host list. */
  private final HostList hostlist;
  /** Initializes the receiver thread with the link to console and host list.
   *  Also starts the thread.
   *  @param      console link to console.
   *  @param      hostlist link to hostlist.
   */
  Receiver(Console console, HostList hostlist) {
    this.console = console;
    this.hostlist = hostlist;
    setDaemon(true);
    setName("Receiver");
    start();
  }
  /** Process incoming message.
   *  @param      in input stream.
   *  @param      out output stream.
   */
  void process(BufferedReader in, PrintWriter out) throws IOException {
    MessageParser message = new MessageParser(in.readLine());
    if (message.name.equals("MESSAGE")) {
      String nick = message.string();
      String host = message.string();
      int port = message.integer();
      String text = message.remainder();
      hostlist.add(nick, host, port);
      console.writeLine(nick+" : "+text);
      out.println("OK");
    }
    else if (message.name.equals("CONNECT")) {
      hostlist.add(message.string(), message.string(), message.integer());
      out.println("OK");
    }
    else if (message.name.equals("DISCONNECT")) {
      hostlist.delete(message.string());
      out.println("OK");
    }
    else if (message.name.equals("GET")) {
      out.println("ITEM "+hostlist.me);
      Iterator i = hostlist.iterator();
      while (i.hasNext()) out.println("ITEM "+i.next());
      out.println("OK");
    }
    else if (message.name.equals("FILE")) {
      String nick = message.string();
      String host = message.string();
      int port = message.integer();
      int size = message.integer();
      String name = message.remainder();
      hostlist.add(nick, host, port);
      console.writeLine(nick+" sends file "+name+" ("+size+" B)");
      FileTransfer.decode(new File(name).getName(),in,size);
      console.writeLine("File received succesfully");
      out.println("OK");
    }
    else {
      throw new ProtocolException("Unknown message \""+message.name+"\"");
    }
  }
  /** Receiver thread's body. Waits in the loop for either a connection or
   *  the change in listening port setup.
   */
  public void run() {
    try {
      while (true) {
        while (hostlist.me==null) sleep(100);
        HostList.HostItem actual = hostlist.me;
        console.setTitle(actual.nick+" at "+actual.host+":"+actual.port);
        ServerSocket listening = new ServerSocket(actual.port, 5);
        listening.setSoTimeout(100);
        while (true) {
          if (hostlist.me!=actual) break;
          try {
            Socket accepted = listening.accept();
            accepted.setSoTimeout(3000);
            BufferedReader in = new BufferedReader(new InputStreamReader(accepted.getInputStream(),"UTF-8"));
            PrintWriter out = new PrintWriter(new OutputStreamWriter(accepted.getOutputStream(),"UTF-8"),true);
            process(in, out);
            in.close();
            out.close();
          }
          catch (InterruptedIOException e) {
          }
          catch (IOException e) {
            console.error("ERROR IN RECEIVER THREAD (" + e + ")");
          }
        }
        listening.close();
      }
    }
    catch (IOException e) {
      console.error("FATAL ERROR IN RECEIVER THREAD (" + e + ")");
    }
    catch (InterruptedException e) {
    }
  }
}

/**
 * A sender thread. Reads commands from the console and processes
 * them sychronously, creating connections to other machines and
 * updating <code>Console</code> and <code>HostList</code> objects.
 *
 * @see Receiver
 * @see Console
 * @see HostList
 */
class Sender extends Thread {
  /** The link to common console. */
  private final Console console;
  /** The link to common host list. */
  private final HostList hostlist;
  /** Initializes the sender thread with the link to console and host list.
   *  Also starts the thread.
   *  @param      console link to console.
   *  @param      hostlist link to hostlist.
   */
  Sender(Console console, HostList hostlist) {
    this.console = console;
    this.hostlist = hostlist;
    setName("Sender");
    start();
  }
  /** Send a prepared message to one remote user.
   *  @param      item remote host identification.
   *  @param      message a string to be sent.
   */
  private void send(HostList.HostItem item, String message) throws IOException {
    try {
      Socket sending = new Socket(item.host,item.port);
      sending.setSoTimeout(3000);
      BufferedReader in = new BufferedReader(new InputStreamReader(sending.getInputStream()));
      PrintWriter out = new PrintWriter(new OutputStreamWriter(sending.getOutputStream(),"UTF-8"),true);
      out.println(message);
      if (!in.readLine().equals("OK")) throw new ProtocolException("Not acknowledged");
      in.close();
      out.close();
    }
    catch (IOException e) {
      hostlist.delete(item.nick);
      console.error(item.nick+" unreachable (" + e + ")");
    }
  }
  /** Send a prepared message to all users from the host list (excluding myself).
   *  @param      message a string to be sent.
   */
  private void send2all(String message) throws IOException {
    Iterator i = hostlist.iterator();
    while (i.hasNext()) {
      send((HostList.HostItem)i.next(),message);
    }
  }
  /** Registers this host on the server, and tries to connect to one 
   *  of the hosts on the list returned.
   *  @param      script server URL.
   *  @param      room a name of the room.
   */
  private void connect(String script, String room) throws IOException {
    if (hostlist.me==null) {
      console.error("Identity not set up");
      return;
    }
    try {
      URL url = new URL(script+"?"+URLEncoder.encode(room+" "+hostlist.me));
      URLConnection connection = url.openConnection();
      BufferedReader in = new BufferedReader(new InputStreamReader(connection.getInputStream()));
      String line;
      boolean success = false;
      while ((line = in.readLine())!=null) {
        MessageParser message = new MessageParser(line);
        if (!message.name.equals("ITEM")) throw new ProtocolException("Wrong data format");
        String nick = message.string();
        String host = message.string();
        int port = message.integer();
        if (!nick.equalsIgnoreCase(hostlist.me.nick)) {
          if (connect(host,port,true)) {
            success = true;
            console.writeLine("CONNECTED VIA "+nick+" at "+host+":"+port);
            break;
          }
        }
      }
      in.close();
      if (!success) console.error("NO CONNECTION MADE, WAITING");
    }
    catch (IOException e) {
      hostlist.clear();
      console.error("Connect failed (" + e + ")");
    }
  }
  /** Connects to host specified and fills the hostlist with the data returned.
   *  @param      host host IP or host name.
   *  @param      port host port.
   *  @param      silent don't print errors, just return the result.
   *  @return     true when successful.
   */
  private boolean connect(String host, int port, boolean silent) throws IOException {
    if (hostlist.me==null) {
      console.error("Identity not set up");
      return false;
    }
    try {
      hostlist.clear();
      Socket sending = new Socket(host,port);
      sending.setSoTimeout(3000);
      BufferedReader in = new BufferedReader(new InputStreamReader(sending.getInputStream()));
      PrintWriter out = new PrintWriter(new OutputStreamWriter(sending.getOutputStream(),"UTF-8"),true);
      out.println("GET");
      String line;
      while (!(line = in.readLine()).equals("OK")) {
        MessageParser message = new MessageParser(line);
        if (!message.name.equals("ITEM")) throw new ProtocolException("Wrong data format");
        hostlist.add(message.string(),message.string(),message.integer());
      }
      in.close();
      out.close();
      send2all("CONNECT "+hostlist.me);
    }
    catch (IOException e) {
      hostlist.clear();
      if (!silent) console.error("Connect failed (" + e + ")");
      return false;
    }
    return true;
  }
  /** Disconnects from the network. This is done by sending out a
   *  DISCONNECT message and clearing the hostlist.
   */
  private void disconnect() throws IOException {
    if (hostlist.me!=null) send2all("DISCONNECT "+hostlist.me);
    hostlist.clear();
  }
  /** Sends a chat message to all the users, including myself.
   *  @param      text a message to be send.
   */
  private void message(String text) throws IOException {
    if (hostlist.me==null) {
      console.error("Identity not set up");
    }
    else {
      send2all("MESSAGE "+hostlist.me+" "+text);
      send(hostlist.me,"MESSAGE "+hostlist.me+" "+text);
    }
  }
  /** Sends a file to all the users from the hostlist, but not myself.
   *  @param      name a name of the file.
   */
  private void file(String name) throws IOException {
    console.writeLine("Encoding file");
    StringWriter buffer = new StringWriter();
    PrintWriter writer = new PrintWriter(buffer);
    long len = new File(name).length();
    writer.println("FILE "+hostlist.me+" "+len+" "+name);
    FileTransfer.encode(name,writer);
    writer.flush();
    console.writeLine("Sending file ("+len+" B)");
    send2all(buffer.toString());
    console.writeLine("File sent successfully");
  }
  /** Process a command from the console.
   *  @param      line the command.
   *  @return     true if the sender thread should be terminated.
   */
  private boolean process(String line) throws IOException {
    MessageParser message = new MessageParser(line);
    if (message.name.equalsIgnoreCase("CONNECT")) {
      connect(message.string(),message.integer(),false);
    }
    else if (message.name.equalsIgnoreCase("ROOM")) {
      connect(message.url(),message.string());
    }
    else if (message.name.equalsIgnoreCase("SETUP")) {
      disconnect();
      hostlist.setIdentity(message.string(),message.string(),message.integer());
    }
    else if (message.name.equalsIgnoreCase("DISCONNECT")) {
      disconnect();
    }
    else if (message.name.equalsIgnoreCase("FILE")) {
      file(message.remainder());
    }
    else if (message.name.equalsIgnoreCase("EXIT")) {
      disconnect();
      return false;
    }
    else if (!line.equals("")) {
      message(line);
    }
    return true;
  }
  /** Process a command from the console and handle caught exceptions
   *  @param      line the command.
   *  @return     true if the sender thread should be terminated.
   */
  public boolean processHandled(String line) {
    try {
      if (!process(line)) return false;
    }
    catch (IOException e) {
      console.error("ERROR IN SENDER THREAD (" + e + ")");
    }
    return true;
  }
  /** Sender thread's body. Waits in the loop for commands from the console.
   */
  public void run() {
    while (processHandled(console.readLine())) { }
    console.close();
  }
}

/**
 * A main application class with several static methods used to start
 * the application.
 */
public class Chat {
  /** A default local host IP, resolved at startup. */
  public static final String defaultHost;
  /** A default local listening port, set to 6060. */
  public static final int defaultPort = 6060;
  /** A default URL for server directed connection. */
  public static final String defaultURL = "http://moon.felk.cvut.cz/~valeko/chat/chat.cgi";
  /** Initializes <code>defaultHost</code> with resolved value. */
  static {
    String localhost;
    try { localhost = InetAddress.getLocalHost().getHostAddress(); }
    catch (UnknownHostException e) { localhost = "127.0.0.1"; }
    defaultHost = localhost;
  }
  /** Create a new chat instance with no identity set up.
   *  @param      console a console to use.
   */
  public static void CreateChat(Console console) {
    HostList hostlist = new HostList(console);
    new Receiver(console,hostlist);
    new Sender(console,hostlist);
  }
  /** Create a new chat instance with given identity.
   *  @param     console a console to use.
   *  @param     nick user nickname
   *  @param     host user hostname (IP adress)
   *  @param     port user port
   */
  public static void CreateChat(Console console, String nick, String host, int port) {
    HostList hostlist = new HostList(console);
    hostlist.setIdentity(nick, host, port);
    new Receiver(console,hostlist);
    new Sender(console,hostlist);
  }
  /** Create a new chat instance with given identity and connect to network via server.
   *  @param     console a console to use.
   *  @param     nick user nickname
   *  @param     host user hostname (IP adress)
   *  @param     port user port
   *  @param     url contact server script URL
   *  @param     room name of the room
   */
  public static void CreateChat(Console console, String nick, String host, int port, String url, String room) {
    HostList hostlist = new HostList(console);
    hostlist.setIdentity(nick, host, port);
    new Receiver(console,hostlist);
    new Sender(console,hostlist).processHandled("ROOM "+url+" "+room);
  }
  /** Starts and initializes application according to command line arguments.
   *  @param     args command line arguments.
   */
  public static void main(String args[]) {
    try {
      String nick = null;
      String host = defaultHost;
      int port = defaultPort;
      String url = null;
      String room = null;
      boolean text = false;
      for (int i=0; i<args.length; i++) {
        if (args[i].equalsIgnoreCase("-text")) {
          text = true;
        }
        else if (args[i].equalsIgnoreCase("-nick")) {
          nick = args[++i];
        }
        else if (args[i].equalsIgnoreCase("-setup")) {
          nick = args[++i];
          host = args[++i];
          port = Integer.parseInt(args[++i]);
        }
        else if (args[i].equalsIgnoreCase("-room")) {
          url = args[++i];
          room = args[++i];
        }
      }
      Console console = text ? (Console)new TextConsole() : (Console)new AWTConsole();
      if (nick==null) CreateChat(console);
      else if (url==null) CreateChat(console, nick, host, port);
      else CreateChat(console, nick, host, port, url, room);
    }
    catch (Exception e) {
      System.err.println("STARTUP ERROR (" + e + ")");
    }
  }
}
